class Entity:
    def __init__(self, name='Entity'):
        self.is_initialized = False
        self.is_active = False
        self.parts = {}
        self.parts_to_add = []
        self.parts_to_remove = []
        self.name = name
        self.entity_manager = None

    def has(self, part):
        return part in self.parts.keys()

    def get(self, part_cls):
        if not self.has(part_cls):
            raise ValueError('Entity {} does not contain {} part.'.format(self.name, part_cls.__name__))
        return self.parts[part_cls]

    def attach(self, part):
        cls = type(part)
        if self.has(cls):
            raise ValueError('Entity {} already contain part {}.'.format(self.name, cls.__name__))

        self.parts[cls] = part
        part.entity = self
        if self.is_initialized:
            part.initialize()

    def replace(self, part):
        if self.has(part.__class__):
            self.detach(part.__class__)
        if self.is_initialized:
            self.parts_to_add.append(part)
        else:
            self.attach(part)

    def detach(self, part_cls):
        if self.has(part_cls) and part_cls not in self.parts_to_remove:
            self.parts_to_remove.append(part_cls)

    def initialize(self):
        self.is_initialized = True
        self.is_active = True
        for p in self.parts.values():
            p.initialize()

    def cleanup(self):
        self.is_active = False
        for p in self.parts.values():
            p.cleanup()

    def update(self, delta):
        for p in self.parts.values():
            if p.is_active:
                p.update(delta)
        while self.parts_to_remove:
            self.remove(self.parts_to_remove.pop(0))

        while self.parts_to_add:
            self.attach(self.parts_to_add.pop(0))

    def remove(self, part_cls):
        if not self.has(part_cls):
            raise ValueError('Entity does not contain {} part.'.format(str(part_cls)))
        self.parts[part_cls].cleanup()
        del self.parts[part_cls]

    def on_collision_with(self, other_collider):
        for key, value in self.parts.items():
            collision_callback = getattr(value, "on_collision_with", None)
            if callable(collision_callback):
                collision_callback(other_collider)