from Core.Entity import Entity
from Core.EventManager import Event
from Core.Parts.HealthPart import HealthPart
from Core.Parts.TextPart import TextPart


class UISystem:
    def __init__(self, entity_manager, event_manager, background):
        self.background = background
        self.event_manager = event_manager
        self.texts = {}
        self.hp_values = None

        self.entity_manager = entity_manager

        self.event_manager.register(Event.ENTITY_TOOK_HIT, self.on_take_hit)

        self.select_mode_solo_text = None
        self.select_mode_pvp_text = None
        self.select_mode_controls_text = None
        self.game_over_text = None

    def initialize_players(self, player, enemy):
        self.hp_values = {
            player.name: player.get(HealthPart).max_health,
            enemy.name: enemy.get(HealthPart).max_health
        }

        player_text_entity = self.create_text(self.health_info_str(player.name, self.hp_values[player.name]), (0, 0), ('left', 'top'))
        enemy_text_entity = self.create_text(self.health_info_str(enemy.name, self.hp_values[enemy.name]), (800, 0), ('right', 'top'))
        enemy_text_entity.top_right = self.background.get_width()

        self.texts[player.name] = player_text_entity.get(TextPart)
        self.texts[enemy.name] = enemy_text_entity.get(TextPart)

        self.entity_manager.add(player_text_entity)
        self.entity_manager.add(enemy_text_entity)

    def on_take_hit(self, entity):
        health_part = entity.get(HealthPart)
        self.hp_values[entity.name] = health_part.health
        self.texts[entity.name].set_text(self.health_info_str(entity.name, health_part.health))

    def show_game_over(self, who_died):
        self.game_over_text = UISystem.create_game_over_text(who_died, (0, 0), ('center', 'center'))
        self.entity_manager.add(self.game_over_text)

    def hide_game_over(self):
        self.entity_manager.remove(self.game_over_text)
        self.game_over_text = None

    def show_select_mode(self):
        self.select_mode_solo_text = UISystem.create_text('Press 1 to play against computer', (0, -50), ('center', 'center'))
        self.select_mode_pvp_text = UISystem.create_text('Press 2 to play against other player', (0, -25), ('center', 'center'))
        self.select_mode_controls_text = UISystem.create_text('Controls: check README', (0, 50), ('center', 'center'))
        self.entity_manager.add(self.select_mode_solo_text)
        self.entity_manager.add(self.select_mode_pvp_text)
        self.entity_manager.add(self.select_mode_controls_text)

    def hide_select_mode(self):
        self.entity_manager.remove(self.select_mode_solo_text)
        self.entity_manager.remove(self.select_mode_pvp_text)
        self.entity_manager.remove(self.select_mode_controls_text)
        self.select_mode_solo_text = None
        self.select_mode_pvp_text = None
        self.select_mode_controls_text = None

    def clear(self):
        self.texts = {}
        self.hp_values = None

    @staticmethod
    def health_info_str(name, hp):
        return '{} health: {}'.format(name, max(0, hp))

    @staticmethod
    def create_text(msg, pos, alignment):
        entity = Entity()
        entity.attach(TextPart(msg, pos=pos, alignment=alignment))
        return entity

    @staticmethod
    def create_game_over_text(entity_name, pos, alignment):
        msg = 'Game Over! {} destroyed! Press ENTER to restart!'.format(entity_name)
        entity = UISystem.create_text(msg, pos, alignment)
        entity.is_active = False
        return entity
