from enum import Enum

import ResourceUtils
from Core.Entity import Entity
from Core.EntityManager import EntityManager
from Core.EventManager import EventManager, Event
from Core.MathUtils import random_sign
from Core.Parts.AudioSourcePart import AudioSourcePart
from Core.Parts.BoxColliderPart import BoxColliderPart
from Core.Parts.CircleColliderPart import CircleColliderPart
from Core.Parts.EnemyAIPart import EnemyAIPart
from Core.Parts.HealthPart import HealthPart
from Core.Parts.MissileLauncherPart import MissileLauncherPart
from Core.Parts.OnCircleMovementPart import OnCircleMovementPart
from Core.Parts.PlanetPart import PlanetPart
from Core.Parts.PlayerVesselInputPart import PlayerVesselInputPart
from Core.Parts.SpritePart import SpritePart
from Core.Parts.TransformPart import TransformPart
from Core.Parts.VesselPart import VesselPart
from Core.Systems.CollisionSystem import CollisionSystem, CollisionLayer
from Core.Systems.InputManager import InputManager, InputType
from Core.Systems.UISystem import UISystem
from ResourceUtils import *
from pygame.locals import QUIT, KEYUP, K_ESCAPE
from random import Random, random

rand = Random()

'''
TODO:
    - bug with input, mash A/D and input hands at some point
    - enemy engine sound stops at some point
'''


def create_planets(planet_size, screen_size):
    # even distribution of planets on screen

    # fill factor of the planet,
    # multiplying it by some value to make them spread a little
    fp = planet_size[0] / screen_size[0] * 4

    # fill factor of the empty space
    ff = (1 - 2 * fp) / 3

    planet1_pos_x = (ff + fp / 2) * screen_size[0]
    planet2_pos_x = (2 * ff + 1.5 * fp) * screen_size[0]

    player1_planet = create_planet('Player1Planet', (planet1_pos_x, screen_size[1] / 2), 'planet.png')
    planet2_planet = create_planet('Player2Planet', (planet2_pos_x, screen_size[1] / 2), 'planet2.png')

    return player1_planet, planet2_planet


def create_vessels(planets, entity_manager):
    player1_vessel = create_vessel('Player1', planets[0], entity_manager)
    player2_vessel = create_vessel('Player2', planets[1], entity_manager)
    return player1_vessel, player2_vessel


def create_planet(name, position, sprite):
    planet = create_base_game_object(name, sprite, CollisionLayer.Planet, use_circ_col=True)
    planet.get(TransformPart).move(position)
    planet.attach(PlanetPart(20 + random() * 20, random_sign()))
    return planet


def create_vessel(name, planet, entity_manager):
    vessel = create_base_game_object(name, 'vessel.png', CollisionLayer.Vessel, flipped_sprite='vessel_flipped.png')
    vessel.attach(OnCircleMovementPart(planet.get(PlanetPart), 105))
    engine_audio = AudioSourcePart(ResourceUtils.load_sound('vessel_engine.ogg'))
    vessel.attach(engine_audio)
    vessel.attach(VesselPart(2, {'engine_audio': engine_audio}))
    vessel.attach(HealthPart(100))
    vessel.attach(MissileLauncherPart(entity_manager))
    return vessel


def create_base_game_object(name, sprite_name, collision_layer, use_circ_col=False, flipped_sprite=None):
    char = Entity(name)

    flipped_sprite_loaded = None
    if flipped_sprite:
        flipped_sprite_loaded = load_image(flipped_sprite, -1)

    char.attach(SpritePart(load_image(sprite_name, -1), flipped_sprite_loaded))
    char.attach(TransformPart())

    if use_circ_col:
        char.attach(CircleColliderPart(collision_layer))
    else:
        char.attach(BoxColliderPart(collision_layer))
    return char


class GameState:

    class GameMode(Enum):
        SOLO = 1
        PVP = 2

    class GameState(Enum):
        UNSETTLED = 1
        MODE_SELECTION = 2
        PLAYING = 3
        GAME_OVER = 4

    def __init__(self, input_manager, event_manager, entity_manager, ui_manager, screen):
        self.should_finish = False

        self.player1 = None
        self.player2 = None
        self.player_dead = None

        self.actor_names = None

        self.mode_selected = None
        self.ready_to_restart = False

        self.game_state = GameState.GameState.UNSETTLED

        self.input_manager = input_manager
        self.event_manager = event_manager
        self.ui_manager = ui_manager
        self.entity_manager = entity_manager

        self.screen = screen

        input_manager.register(InputType.BUTTON, self.on_button_down)
        event_manager.register(Event.ENTITY_DEAD, self.on_entity_dead)

    def clear(self):
        self.entity_manager.clear()
        self.ui_manager.clear()
        self.input_manager.clear()
        self.input_manager.register(InputType.BUTTON, self.on_button_down)

        self.player1 = None
        self.player2 = None
        self.actor_names = None

        self.mode_selected = None
        self.ready_to_restart = False

        self.player_dead = None

    def create_game(self):
        planets = create_planets((128, 128), self.screen.get_size())
        self.player1, self.player2 = create_vessels(planets, self.entity_manager)

        self.player1.get(MissileLauncherPart).target = self.player2.get(TransformPart)
        self.player2.get(MissileLauncherPart).target = self.player1.get(TransformPart)

        self.player1.attach(PlayerVesselInputPart(self.input_manager, {'left': K_a, 'right': K_d, 'fire': K_SPACE}))

        if self.mode_selected == GameState.GameMode.SOLO:
            self.player2.name = 'Computer'
            self.player2.attach(EnemyAIPart())
        else:
            self.player2.attach(PlayerVesselInputPart(self.input_manager, {'left': K_LEFT, 'right': K_RIGHT, 'fire': K_RCTRL}))
            self.player2.name = 'Player2'

        self.actor_names = (self.player1.name, self.player2.name)
        self.entity_manager.add_all((self.player1, self.player2))
        self.entity_manager.add_all(planets)

        self.ui_manager.initialize_players(self.player1, self.player2)

        self.ui_manager.hide_select_mode()

    def update(self):
        if self.game_state == GameState.GameState.UNSETTLED:
            self.ui_manager.show_select_mode()
            self.game_state = GameState.GameState.MODE_SELECTION
        elif self.game_state == GameState.GameState.MODE_SELECTION:
            if self.mode_selected:
                self.game_state = GameState.GameState.PLAYING
                self.create_game()
        elif self.game_state == GameState.GameState.PLAYING:
            if self.player_dead:
                self.game_state = GameState.GameState.GAME_OVER
                self.ui_manager.show_game_over(self.player_dead.name)
                self.clear()
        elif self.game_state == GameState.GameState.GAME_OVER:
            if self.ready_to_restart:
                self.game_state = GameState.GameState.UNSETTLED
                self.ui_manager.hide_game_over()

    def on_entity_dead(self, entity):
        if entity.name in self.actor_names:
            self.player_dead = entity

    def restart(self):
        self.clear()

    def on_button_down(self, *args):
        if len(args) == 2:
            if args[1] == KEYUP:
                if args[0] == K_ESCAPE:
                    self.should_finish = True
                elif args[0] == K_RETURN and self.game_state == GameState.GameState.GAME_OVER:
                    self.ready_to_restart = True
                elif self.game_state == GameState.GameState.MODE_SELECTION:
                    if args[0] == K_1:
                        self.on_mode_selected(GameState.GameMode.SOLO)
                    elif args[0] == K_2:
                        self.on_mode_selected(GameState.GameMode.PVP)
        elif args[0] == QUIT:
            self.should_finish = True

    def on_mode_selected(self, mode):
        self.mode_selected = mode


def main():
    # Initialize Everything

    # the larger the buffer, the more lag sound seems to have
    pygame.mixer.pre_init(44100, -16, 1, 4096)
    pygame.init()

    validate_resources()

    screen = pygame.display.set_mode((800, 600))
    pygame.display.set_caption('Planet Duel')
    pygame.mouse.set_visible(0)

    # Create The Background
    background = pygame.Surface(screen.get_size())
    background = background.convert()
    background.fill((16, 16, 16))

    # Prepare Game Objects
    clock = pygame.time.Clock()

    # Create systems
    event_manager = EventManager()
    input_manager = InputManager()
    entity_manager = EntityManager(event_manager, screen, background)
    collision_system = CollisionSystem(entity_manager)
    ui_system = UISystem(entity_manager, event_manager, background)

    game_state = GameState(input_manager, event_manager, entity_manager, ui_system, screen)

    while not game_state.should_finish:
        millis = clock.tick(60)

        game_state.update()
        input_manager.update()

        entity_manager.update(millis / 1000)

        collision_system.update()

        if not entity_manager.layered_dirty:
            entity_manager.create_layered_dirty()
        else:
            entity_manager.layered_dirty.update()
            rects = entity_manager.layered_dirty.draw(screen)
            pygame.display.update(rects)

    pygame.quit()


if __name__ == '__main__':
    main()
